//
//================================================================================
//
//  Pro Memoria
//    |_main.cpp
//    |_pm.h
//    |_pm.cpp
//
//================================================================================
//
//  Created by Paul Ochs on 29.06.19.
//
//  A file to remember syntax and features. Not meant as a sensible program.
//  To try out regions of code set them to 1 and all others to 0.
//
//  Some general websites:
//  https://lec.inf.ethz.ch/math/informatik_cse/2022/guides/style/intro.html
//  http://isocpp.org/
//
//================================================================================
//

#include "pm.h"
#include <cassert>
#include <chrono>
#include <future>
#include <iomanip>
#include <iostream>
#include <limits>
#include <random>
#include <vector>

using matrix = std::vector<std::vector<int>>;

template <typename type> type example(type key) { return 2 * key; }

template <typename T, typename Function> void filter(const T &someset, Function f)
{
    // now this is a function, that will work correctly, if the first argument offers an iterator and the second
    // argument is something that can be applied to elements
    for (const auto &x : someset)
    {
        // also the result of f(x) needs to be convertible to bool in some way
        if (f(x))
            std::cout << x << " ";
    }
    std::cout << std::endl;
}

template <typename type>
// also possible to have several types
// template<typename Function, typename Delta> for example
struct functor
{
    // these member variables are then the things you can use as an übergabeparameter
    // i.e. they are the objects you can call the functor with
    type value;
    // Function function; //i.e. making a new object called "function" of type "Function"
    // Delta delta;

    // default constructor here,
    functor(type x) : value(x) {}
    // derivative(Function f, Delta d) : function(f), delta(d) {} //example with more than one templated type

    // overload the () operator, then you can call the functor via this
    bool operator()(type num) const
    {
        // type num means if functor is called with () then it can take one call-parameter
        // we could also add another e.g. int a and then just print it out or sth.
        // std::cout << a;
        return num > value;
    }
};

bool task4(unsigned int a, unsigned int b, unsigned int c) { return (!a && !b) ^ c; }

int main()
{
    std::cout << " \n                 -------------------\n                 |   Pro Memoria   | \n                 "
                 "------------------- \n "
              << "A collection of explanations, links and examples to help remember C++ things." << std::endl;

    // std::cerr << " Print all error messages to std::cerr, all normal output to std::cout" << std::endl;

//================================================================================
// ^ (binary or) vs != (logical not equal) for bool or ints:
//================================================================================
#if 0
    bool ba = true;
    bool bb = false;
    int ia = 5;
    int ib = 0;
    // std::cout << (ba != bb) << std::endl;
    // std::cout << (ba ^ bb) << std::endl;
    // std::cout << (ia != ib) << std::endl;
    // std::cout << (ia ^ ib) << std::endl;
    std::cout << task4(0, 0, 1) << std::endl;
    std::cout << task4(0, 0, 5) << std::endl;
#endif

//================================================================================
// value of assignment statement
//================================================================================
// https://stackoverflow.com/questions/6807393/using-assignment-as-a-condition-expression

//================================================================================
// modulo for negative numbers:
//================================================================================
#if 0
    std::cout << std::numeric_limits<unsigned int>::max() << std::endl;

    std::cout << "-------------------- line\n";
    std::cout << (-7 % 3) << std::endl;
    std::cout << (7 % -3) << std::endl;
    std::cout << (-7 % -3) << std::endl;
    // cout << 5 % -4 << "\n"; //% übernimmt vorzeichen vom linken operanden
#endif

//================================================================================
// range based for loops over container:
//================================================================================
#if 0
    std::vector<int> v(3);
    // vector will be 111
    for (auto &x : v)
        x = 1; // does actually change the vector as we are operating on &x
    std::cout << " vector print 1: ";
    // element based iteration: see "range based for loop"
    for (auto x : v)
        std::cout << x;

    std::cout << "vector print 2: ";
    // same as print 1, but now doing it index based
    for (int i = 0; i < v.size(); i++)
        std::cout << v[i];

    std::cout << " vector edit 1: " << std::endl;
    for (auto i : v)
        ++v[i]; 
        // careful: i is still an element in each iteration
        // it will go to the index that is the element, which works here as we only have 1 in vector
        // at beginning and then after we do ++v[1] i.e. now v[1] == 2, now i == 2 wich is the third element
        // in vector then we increment this. Then v[2] == 2. In the final loop iteration we have i == 2
        // so v[2] is incremented again resulting in cout // 123

    std::cout << " vector edit 2: ";
    for (auto x : v)
        x = 5; // does not change the actual vector as we are not operating on the memory region
    std::cout << " 5 vector print: ";
    for (auto x : v)
        std::cout << x;
    std::cout << " 6 vector print: ";
    for (auto &x : v)
        std::cout << x;
#endif

//================================================================================
// pointers and references:
//================================================================================
#if 0
    std::cout << "pointers and references:" << std::endl;
    const int &constref = 5;
    std::cout << constref << std::endl;
    // int & nonconstref = 5;   
    // error: non-const lvalue reference to type 'int' cannot bind to a temporary of type 'int'

    int x = 1;  // normal variable
    int &y = x; // like an alias for x, same address but called y
    y += 2;     // were doing something with y, but y is just another name for the adresse of x
                // so were doing something to the adresse of x and y and both change
    std::cout << x << ", " << y << std::endl;

    int *z = &x; // z is ptr to x resp the adresse where x lives
    int *w = &y; // w is ptr to y i.e. to x

    // x,y,*z and *w all are the same address i.e. the same value is saved in them
    std::cout << "x:" << x << ", &x:" << &x << std::endl;
    std::cout << "y:" << y << ", &y:" << &y << std::endl;
    std::cout << "z:" << z << ", *z:" << *z << std::endl;
    std::cout << "w:" << w << ", *w:" << *w << std::endl;

#endif

//================================================================================
// pointers and references w.r.t. std::vector class
//================================================================================
#if 0
    std::vector<int> ref_vec(3);
    // vector will be 111
    for (auto &x : ref_vec)
        x = 1; // does actually change the vector as we are operating on &x
    std::cout << " ref vector print 1: ";
    // element based iteration: see "range based for loop"
    for (auto x : ref_vec)
        std::cout << x;
    std::cout << std::endl;

    std::cout << "&ref_vec: " << &ref_vec << std::endl;
    std::cout << "&ref_vec[0]: " << &ref_vec[0] << std::endl;
    std::cout << "ref_vec.data(): " << ref_vec.data() << std::endl;

#endif

//================================================================================
// pre and post increment (and pointers)
//================================================================================
#if 0
    std::cout << "pre and post increment (and pointers)" << std::endl;
    // https://stackoverflow.com/questions/55031546/differences-in-c-source-source-source

    int a4[] = {1, 2, 3, 4};
    int *pa4 = a4;

#endif

//================================================================================
// arrays
//================================================================================
#if 0
    std::cout << "arrays" << std::endl;
    // initializes an array of size 5 with the listed elements
    // is initialized on stack
    int array[] = {1, 2, 3, 4, 5};
    // other way to initialize an array with elements
    // dynamic memory allocation on
    int *array_name = new int[5]{1, 2, 3, 4, 5};

    std::cout << "array, type of array: " << array << ", " << typeid(array).name() << std::endl;
    // outputs address of array and A5_i: array with 5 elements of type int
    std::cout << "array_name, type of array_name: " << array_name << ", " << typeid(array_name).name() << std::endl;
    // outputs address of array_name and Pi: pointer to variable of type int

    std::cout << std::endl;
#endif

//================================================================================
// dynamic memory allocation, new, delete
//================================================================================
#if 0
    // when stuff is initialized to zero:
    // https://stackoverflow.com/questions/7546620/operator-new-initializes-memory-to-zero

    // error or difference between delete and delete[]
    // https://stackoverflow.com/a/6953502/17560853
    
#endif

//================================================================================
// iterators
//================================================================================
#if 0
    std::vector<int> it_vec(3);
    // vector will be 111
    int num = 0;
    for (auto &x : it_vec)
        x = num++;

    // printf("%p \n", &it_vec[0]);
    // std::cout << (*it_vec.begin()) << std::endl;

    for (std::vector<int>::iterator i = it_vec.begin(); i < it_vec.end(); ++i)
    {
        // only until i < vec.end() !
        // end() is "past-the-end" iterator
        std::cout << (*i) << ", ";
    }

#endif

//================================================================================
// where temporary objects are stored
//================================================================================
// https://www.quora.com/In-C-when-a-temporary-object-is-created-where-is-it-stored


//================================================================================
// uniform/list initialization
//================================================================================
// https://stackoverflow.com/questions/30518780/c11-uniform-initialization-for-constructor-and-operator
// https://www.geeksforgeeks.org/uniform-initialization-in-c/
// https://stackoverflow.com/questions/39487065/what-does-return-statement-mean-in-c11


//================================================================================
// templates
//================================================================================
// https://www.codeproject.com/Articles/257589/An-Idiots-Guide-to-Cplusplus-Templates-Part-1#ClassTempl


//================================================================================
// traits (and templates)
//================================================================================
// https://www.internalpointers.com/post/quick-primer-type-traits-modern-cpp


//================================================================================
// Functors and lambdas
//================================================================================
#if 0
    std::cout << "Functors and lambdas" << std::endl;

    std::cout << example<int>(x);
    std::cout << "Functor:" << std::endl;

    std::vector<int> a{1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 16, 19};
    int value8 = 8;
    filter(a, functor<int>(value8));

    // this uses a lambda expression, but does the same as the functor
    filter(a, [value8](int x) { return x > value8; });

    std::cout << std::endl;
#endif

//================================================================================
// Polymorphism, see also more in pm.h
//================================================================================
#if 0
    std::cout << "Polymorphism" << std::endl;

    f();
    g();
    h();

#endif

    printf("\n done");
    std::cout << std::endl;
    return 0;
}

/* std::future

 #include <cassert>
 #include <future>
 #include <iostream>
 #include <random>

 int action(int a, int b){
 std::cout << "thread is working hard to produce a result" << std::endl;

 std::this_thread::sleep_for(std::chrono::milliseconds(1));
 //sets this i.e. the current thread to sleep for the duration set in ()
 std::cout << "thread finished, sending result now" << std::endl;
 return a+b;
 }

 int main(){
 int x = 1;
 int y = 1;
 std::future<int> f_name = std::async(action,x,y);
 //just like with std::thread, we hand over (function, args)
 std::cout << "thread generated, waiting for result, " << std::endl;
 int val = f_name.get();
 //.get() is public member of std::future and just lets you return the value that ended up being
 //generated in the future
 //you can use f_name.get() like a variable of its corresponding type:
 std::cout << "value received: " << ++val << std::endl;

 return 0;
 }

 */
