#ifndef pro_mem_h
#define pro_mem_h
#include <iostream>
#include <string>
#include <cstring>

//================================================================================
// On operators
//================================================================================
// has good overview on how the operators are actually evaluated as members or non members
// https://en.cppreference.com/w/cpp/language/operators
//
// stackoverflow about member/non-member operators
// https://stackoverflow.com/questions/2425906/operator-overloading-outside-class


//================================================================================
// For Rule of 3/5/0:
//================================================================================
// https://www.codementor.io/@sandesh87/the-rule-of-five-in-c-1pdgpzb04f


//================================================================================
// On move and copy constructors:
//================================================================================
// https://www.geeksforgeeks.org/move-constructors-in-c-with-examples/
//
// https://www.geeksforgeeks.org/copy-constructor-vs-assignment-operator-in-c/

// on move constructors /semantics and the difference between copy constructors:
// https://stackoverflow.com/questions/64378721/what-is-the-difference-between-the-copy-constructor-and-move-constructor-in-c
// more on move semantics
// https://stackoverflow.com/questions/3106110/what-is-move-semantics

// since this is all related to lvalue references vs. rvalue references:
// https://medium.com/pranayaggarwal25/rvalue-references-e99dfd3933ff


class Student
{
private:
    int id;
    char *name;

public:
    Student()
    {
        this->id = 0;
        this->name = nullptr;
    }

    Student(char *name, int id)
    {
        this->id = id;
        this->name = new char[strlen(name) + 1];
        strcpy(this->name, name);
    }

    ~Student() { delete[] this->name; }

    // copy constructor
    Student(const Student &other)
    {
        this->id = other.id;
        this->name = new char[strlen(other.name) + 1];
        strcpy(this->name, other.name);
    }

    // copy assignment
    Student &operator=(const Student &rhs)
    {
        // self-asignment check
        if (this != &rhs)
        {

            this->id = rhs.id;

            // delete old data
            if (this->name)
            {
                delete[] this->name;
            }

            this->name = new char[strlen(rhs.name) + 1];
            strcpy(this->name, rhs.name);
        }

        return *this;
    }

    // Functions
    int take_five(int);
};

// ============================================================================
// Polymorphism

class A
{
public:
    std::string name;
    A(std::string _name) : name(_name) {}

    // why we need virtual: https://stackoverflow.com/questions/2391679/why-do-we-need-virtual-functions-in-c
    // https://www.geeksforgeeks.org/virtual-function-cpp/
    // One of the key features of class inheritance is that a pointer to a derived class is type-compatible with a
    // pointer to its base class. Polymorphism is the art of taking advantage of this simple but powerful and versatile
    // feature. (https://cplusplus.com/doc/tutorial/polymorphism/)

    // https://stackoverflow.com/questions/7223613/c-polymorphism-without-pointers
    // Polymorphism only works with non-value types: references and pointers. And since references can only be bound
    // once, you cannot really use them in standard containers. That leaves you with pointers.

    virtual void say_hello() { std::cout << "A says hi to " << name << "\n"; }

    void say_bye() { std::cout << "A says bye to " << name << "\n"; }
};

class B : public A
{
public:
    B(std::string _name) : A(_name) {}
    void say_hello() { std::cout << "B greets " << name << "\n"; }
};

class C : public A
{
public:
    int a = 5;

    C(std::string _name) : A(_name) {}
    void say_bye() { std::cout << "C say goodbye to " << name << "\n"; }
};


#endif